package de.kif.frontend.views

import de.kif.common.CALENDAR_GRID_WIDTH
import de.kif.common.model.Room
import de.kif.common.model.Schedule
import de.kif.frontend.iterator
import de.kif.frontend.launch
import de.kif.frontend.repository.RepositoryDelegate
import de.kif.frontend.repository.RoomRepository
import de.kif.frontend.repository.ScheduleRepository
import de.westermann.kwebview.*
import de.westermann.kwebview.components.Body
import org.w3c.dom.HTMLAnchorElement
import org.w3c.dom.HTMLElement
import org.w3c.dom.events.EventListener
import org.w3c.dom.events.MouseEvent
import org.w3c.dom.get
import kotlin.browser.document
import kotlin.dom.appendText
import kotlin.dom.isText

class CalendarTools(entry: CalendarEntry, view: HTMLElement) : View(view) {

    init {
        var linkM10: HTMLAnchorElement? = null
        var linkM5: HTMLAnchorElement? = null
        var linkReset: HTMLAnchorElement? = null
        var linkP5: HTMLAnchorElement? = null
        var linkP10: HTMLAnchorElement? = null
        var linkDel: HTMLAnchorElement? = null

        for (element in html.children) {
            when {
                element.classList.contains("calendar-tools-m10") -> linkM10 = element as? HTMLAnchorElement
                element.classList.contains("calendar-tools-m5") -> linkM5 = element as? HTMLAnchorElement
                element.classList.contains("calendar-tools-reset") -> linkReset = element as? HTMLAnchorElement
                element.classList.contains("calendar-tools-p5") -> linkP5 = element as? HTMLAnchorElement
                element.classList.contains("calendar-tools-p10") -> linkP10 = element as? HTMLAnchorElement
                element.classList.contains("calendar-tools-del") -> linkDel = element as? HTMLAnchorElement
            }
        }

        linkM10 = linkM10 ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-m10")
            link.textContent = "-10"
            html.appendChild(link)
            link
        }
        linkM10.removeAttribute("href")
        linkM10.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                val s = entry.schedule.get()
                ScheduleRepository.update(s.copy(time = s.time - 10))
            }
        })

        linkM5 = linkM5 ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-m5")
            link.textContent = "-5"
            html.appendChild(link)
            link
        }
        linkM5.removeAttribute("href")
        linkM5.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                val s = entry.schedule.get()
                ScheduleRepository.update(s.copy(time = s.time - 5))
            }
        })

        linkReset = linkReset ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-reset")
            link.textContent = "reset"
            html.appendChild(link)
            link
        }
        linkReset.removeAttribute("href")
        linkReset.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                val s = entry.schedule.get()
                ScheduleRepository.update(s.copy(time = s.time / CALENDAR_GRID_WIDTH * CALENDAR_GRID_WIDTH))
            }
        })

        linkP5 = linkP5 ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-p5")
            link.textContent = "+5"
            html.appendChild(link)
            link
        }
        linkP5.removeAttribute("href")
        linkP5.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                val s = entry.schedule.get()
                ScheduleRepository.update(s.copy(time = s.time + 5))
            }
        })

        linkP10 = linkP10 ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-p10")
            link.textContent = "+10"
            html.appendChild(link)
            link
        }
        linkP10.removeAttribute("href")
        linkP10.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                val s = entry.schedule.get()
                ScheduleRepository.update(s.copy(time = s.time + 10))
            }
        })

        linkDel = linkDel ?: run {
            val link = createHtmlView<HTMLAnchorElement>()
            link.classList.add("calendar-tools-del")
            link.textContent = "del"
            html.appendChild(link)
            link
        }
        linkDel.removeAttribute("href")
        linkDel.addEventListener("click", EventListener {
            entry.pending = true
            launch {
                ScheduleRepository.delete(entry.scheduleId)
            }
        })
    }
}

class CalendarEntry(view: HTMLElement) : View(view) {

    private lateinit var mouseDelta: Point
    private var newCell: CalendarCell? = null

    private var language by dataset.property("language")

    val scheduleId = dataset["id"]?.toLongOrNull() ?: 0

    val schedule = RepositoryDelegate(ScheduleRepository, scheduleId)

    var pending by classList.property("pending")

    private fun onMove(event: MouseEvent) {
        val position = event.toPoint() - mouseDelta

        val cell = calendarCells.find {
            position in it.dimension
        }


        if (cell != null) {
            cell += this

            if (newCell == null) {
                style {
                    left = "0"
                    top = "0.1rem"
                }
            }

            newCell = cell
        }

        event.preventDefault()
        event.stopPropagation()
    }

    private fun onFinishMove(event: MouseEvent) {
        classList -= "drag"

        newCell?.let { cell ->
            launch {
                val newTime = cell.time
                val newRoom = cell.getRoom()

                pending = true

                val s = schedule.get().copy(room = newRoom, time = newTime)

                ScheduleRepository.update(s)
            }
        }
        newCell = null

        for (it in listeners) {
            it.detach()
        }
        listeners = emptyList()

        event.preventDefault()
        event.stopPropagation()
    }

    private var listeners: List<de.westermann.kobserve.event.EventListener<*>> = emptyList()

    init {
        onMouseDown { event ->
            if (event.target != html || "pending" in classList) {
                event.stopPropagation()
                return@onMouseDown
            }

            launch {
                classList += "drag"

                val s = schedule.get()
                val time = s.time / CALENDAR_GRID_WIDTH * CALENDAR_GRID_WIDTH

                val p = calendarCells.find {
                    it.day == s.day && it.time == time && it.roomId == s.room.id
                }?.dimension?.center ?: dimension.center

                mouseDelta = event.toPoint() - p

                listeners = listOf(
                    Body.onMouseMove.reference(this::onMove),
                    Body.onMouseUp.reference(this::onFinishMove),
                    Body.onMouseLeave.reference(this::onFinishMove)
                )
            }
            event.preventDefault()
            event.stopPropagation()
        }

        var calendarTools: CalendarTools? = null
        for (item in html.children) {
            if (item.classList.contains("calendar-tools")) {
                calendarTools = CalendarTools(this, item)
                break
            }
        }
        if (calendarTools == null) {
            calendarTools = CalendarTools(this, createHtmlView())
            html.appendChild(calendarTools.html)
        }
    }

    fun load(schedule: Schedule) {
        pending = false

        language = schedule.workGroup.language.code

        this.schedule.set(schedule)

        style {
            val size = schedule.workGroup.length / CALENDAR_GRID_WIDTH.toDouble()
            val pos = (schedule.time % CALENDAR_GRID_WIDTH) / CALENDAR_GRID_WIDTH.toDouble()

            val ps = "${pos * 100}%"
            val sz = "${size * 100}%"

            left = ps
            top = "calc($ps + 0.1rem)"

            width = sz
            height = "calc($sz - 0.2rem)"

            if (schedule.workGroup.track?.color != null) {
                backgroundColor = schedule.workGroup.track.color.toString()
                color = schedule.workGroup.track.color.calcTextColor().toString()
            }
        }

        for (element in html.childNodes) {
            if (element.isText) {
                html.removeChild(element)
            }
        }

        html.appendText(schedule.workGroup.name)


        val time = schedule.time / CALENDAR_GRID_WIDTH * CALENDAR_GRID_WIDTH
        val cell = calendarCells.find {
            it.day == schedule.day && it.time == time && it.roomId == schedule.room.id
        }

        if (cell != null && cell.html != html.parentElement) {
            cell += this
        }
    }

    companion object {
        fun create(schedule: Schedule): CalendarEntry {
            val entry = CalendarEntry(createHtmlView())

            entry.load(schedule)

            return entry
        }
    }
}

class CalendarCell(view: HTMLElement) : ViewCollection<CalendarEntry>(view) {
    val day = dataset["day"]?.toIntOrNull() ?: 0
    val time = dataset["time"]?.toIntOrNull() ?: 0
    val roomId = dataset["room"]?.toLongOrNull() ?: 0

    private lateinit var room: Room

    suspend fun getRoom(): Room {
        if (this::room.isInitialized) {
            return room
        }

        room = RoomRepository.get(roomId) ?: throw NoSuchElementException()
        return room
    }
}

var calendarEntries: List<CalendarEntry> = emptyList()
var calendarCells: List<CalendarCell> = emptyList()

fun initCalendar() {
    calendarEntries = document.getElementsByClassName("calendar-entry")
        .iterator().asSequence().map(::CalendarEntry).toList()

    calendarCells = document.getElementsByClassName("calendar-cell")
        .iterator().asSequence().filter { it.dataset["time"] != null }.map(::CalendarCell).toList()

    ScheduleRepository.onCreate {
        launch {
            val schedule = ScheduleRepository.get(it) ?: throw  NoSuchElementException()
            CalendarEntry.create(schedule)
        }
    }
    ScheduleRepository.onUpdate {
        launch {
            val schedule = ScheduleRepository.get(it) ?: throw  NoSuchElementException()
            var found = false
            for (entry in calendarEntries) {
                if (entry.scheduleId == it) {
                    entry.load(schedule)
                    found = true
                }
            }
            if (!found) {
                CalendarEntry.create(schedule)
            }
        }
    }
    ScheduleRepository.onDelete {
        for (entry in calendarEntries) {
            if (entry.scheduleId == it) {
                entry.html.remove()
            }
        }
    }
}